function [score]=IF_enhancedTesting(forest,test,aggregation,score_types,nTrees,c_l)
% % Enhanced Anomaly Scores for Isolation Forest, v1.0, 2024
% (c) A. Mensi
%
% forest is a trained Isolation Forest
%
% test is the testing set. Each object will traverse each tree within the
% forest.
%
% aggregation is a value indicating how tree scores are combined at forest
% level. The possible values are 'sx' or 'px'.
%
% score_types is a vector that can contain one or more values in the range
% of 0:5 indicating a specific path-weighted variant. 0 corresponds to no
% weight, whereas the remainders correspond to the V1-V5 variants indicated
% in the paper "Enhanced Anomaly scores for Isolation Forests," Pattern
% Recognit., 2021, A. Mensi, M. Bicego
%
% nTrees<=param.T is the number of trees to analyze within the forest
% (i.e., it is  needed if one wanted to test on a smaller portion of the built forest)

%Checking input parameters
param=forest.param;
if nargin<6
    for i=1:param.S
        c_l(i)=average_unsuccessful_search(i);
    end
end
if nargin<5
    nTrees=param.T;
else
    nTrees=min(nTrees,param.T);
end
c_n=c_l(param.S);
if nargin<4
    score_types=0;
elseif ~isempty(setdiff(score_types,0:5))
    [~,idxs]=setdiff(score_types,0:5);
    score_types(idxs)=[];
    if isempty(score_types)
        fprintf("score_types must be one or more values in 0:5. Setting it to 0.\n");
        score_types=0;
    end
end
if nargin<3 ||  ~any(strcmpi(aggregation,{'sx','px'}))
    fprintf("Aggregation must be either 'sx' or 'px'. Setting it to 'sx'.")
        aggregation='sx';
end



nTest=size(test,1);
hlim=param.max_depth;
treeScore=zeros(numel(score_types),nTest,nTrees);
score=zeros(numel(score_types),nTest);

for t=1:nTrees
    tree=forest.tree{t};
    if mod(t,100)==0
        fprintf("Testing iTree #%d\n",t)
    end
    
    %Get additional information from the tree if V4 and V5 need to be
    %computed
    if any(score_types==4) || any(score_types==5) 
        ns=getNodesSize(tree);
        paths = getPaths(tree);
        lcaDepth=getLCA(tree,paths);
    end
    feat_order = tree.feat_idx;
    for n=1:nTest
        %Tree traversal for testing object n
        node_tmp=tree;
        test_ord = test(n,feat_order);
        while ~isempty(node_tmp.left) 
            if test_ord(node_tmp.feat)<node_tmp.value
                node_tmp=node_tmp.left;
            else
                node_tmp=node_tmp.right;
            end
            if any(score_types==2) || any (score_types==3)
                if node_tmp.proxy~=0
                    treeScore(score_types==2,n,t)=treeScore(score_types==2,n,t)+1/node_tmp.proxy;
                    treeScore(score_types==3,n,t)=treeScore(score_types==3,n,t)+((1/node_tmp.proxy)*(1/node_tmp.nsamples));
                end
            end
            if any(score_types==1)
                treeScore(score_types==1,n,t)=treeScore(score_types==1,n,t)+1/node_tmp.nsamples;
            end
            if node_tmp.height==hlim
                break
            end
        end
        %Unweighted score at tree level
        treeStdScore=node_tmp.height; 
        if node_tmp.height==hlim
            treeStdScore=treeStdScore+c_l(node_tmp.nsamples);
            treeScore(ismember(score_types, 1:3),n,t)=treeScore(ismember(score_types, 1:3),n,t)+c_l(node_tmp.nsamples);
        end
        if any(score_types==0)
            treeScore(score_types==0,n,t)=treeStdScore;
        end
        if any(score_types==4) || any(score_types==5)
            treeScore(score_types==4,n,t)=sum((treeStdScore-lcaDepth(:,paths== str2num(node_tmp.path))').*ns)./tree.nsamples;
            treeScore(score_types==5,n,t)=sum(2.^(-treeStdScore+lcaDepth(:,paths== str2num(node_tmp.path))').*ns)./tree.nsamples;
        end
    end
end

%Aggregating scores
if strcmpi(aggregation,'sx')
    score(ismember(score_types, 0:4),:)=2.^(-mean(treeScore(ismember(score_types, 0:4),:,:),3)./c_n);
else %'px'
    score(ismember(score_types, 0:4),:)=mean(2.^-treeScore(ismember(score_types, 0:4),:,:),3)./c_n;
end
score(score_types==5,:)=mean(treeScore(score_types==5,:,:),3)./c_n; %V5 has its own aggregation 
end