function [forest,param]=IF_training(train,param)
% Enhanced Anomaly Scores for Isolation Forest, v1.0, 2024
% (c) A. Mensi
%
% forest=IF_training(train,param) creates an Isolation Forest (IF) of param.T
% Isolation Trees from train. Differently from standard Isolation Trees
% additional information are stored, used during the computation of
% different anomaly scores.
%
% train is a distance matrix NxN containing pairwise distances.
%
% param is a structure containing the following fields:
%   -T: the number of Isolation Trees (iTree) in an IF.
%           DEFAULT: T=100.
%   -S: the number of objects used to build an iTree. The objects are randomly subsampled without replacement
%       from train for each tree in an independent way.
%           DEFAULT: S=256.
%   -D: the maximum depth reachable in each iTree. It can either
%       be 'log', i.e. ceil(log_2(S)), or 'max', i.e. S-1.
%           DEFAULT: 'log'
%   -F: the percentage of features used to build an iTree. ceil(param.F)
%       features are randomly subsampled without replacement from train for
%       each tree in an independent way. The value ranges in (0,1]
%           DEFAULT: 1
%   -gamma: parameter needed to compute the proxy (used in V2 and V3)
%           DEFAULT: 1


%Checking correctness of forest parameters
[N,d]=size(train);
param=setIFParameters(param,N,d);

tree=cell(1,param.T);
for t=1:param.T
    if mod(t,100)==0
        fprintf("Training iTree #%d\n",t)
    end

    %Sampling objects and features
    tree{t}.idx=randsample(N,param.S);
    tree{t}.feat_idx=randsample(d,param.nFeat);
    tr=train(tree{t}.idx,tree{t}.feat_idx);
    tree{t}.height=0;

    %Training
    tree{t}=iTree(tr,tree{t},param);
end
forest.tree=tree;
forest.param=param;
end