function tree=iTree(train,tree,param)
% Enhanced Anomaly Scores for Isolation Forest, v1.0, 2024
% (c) A. Mensi
%
% tree=iTree(train,tree,param) recursive function that builds a ProxIT.
%Implementation required not only to go from root to leaves but also from
%leaves to root
% The function returns a ProxIT tree structure that can be traversed if distances
% to the training objects are known.
%
% train is the original distance matrix containing pairwise distances between all training objects
%
% tree is a nested tree structure where each level corresponds to a node.
% The most external level corresponds to the root node. The structure, at
% each level, contains the following fields:
%   -nsamples: number of objects that have reached the current node.
%   numel(idx) gets the same information.
%   -left (right): pointer to the left (right) child node of the current node.
%   Empty if current node is a leaf.e.
%    -imp: contains the value of the optimization function. Empty if
%    criterion is random or current node is a leaf.
%   -If criterion is 1P, i.e. param.thr=1:
%       -tree.proto contains the index of the prototype defining the test
%       of the current node.
%       -tree.thr contains the index of the threshold  defining the test
%       of the current node.
%   -If criterion is 2P, i.e. param.thr=0:
%       -tree.protoL contains the indx of the left prototype defining the test
%       of the current node.
%       -tree.protoR contains the index of the right prototype  defining the test
%       of the current node.
%
% param is a structure containing several fields which refer to the
% parameters of the ProxIF model. For more information write 'help 'IF_training'
N=size(train,1);
tree.nsamples=N;

if tree.height==0 %root
    tree.volume=lebesgue(train);
    tree.limits=[min(train)' max(train)'];
    tree.path = '1';
end

if N==1 || tree.height==param.max_depth || size(unique(train,'rows'),1)==1 %leaf
    tree.feat=[]; tree.value=[]; tree.left=[]; tree.right=[]; tree.proxy=[];
    tree.isLeaf=1;
else
    tree.isLeaf=0;

    %Test on the current node: selecting both the feature and the threshold at random
    tree.feat=randsample(param.nFeat,1);
    while isscalar(unique(train(:,tree.feat))) %If I pick up a feature which is equal across all objects in the node I pick another one
        tree.feat=randsample(param.nFeat,1);
    end
    a=min(train(:,tree.feat))+realmin; b=max(train(:,tree.feat));
    tr_left=[]; tr_right=[];
    while isempty(tr_left) || isempty(tr_right)
        tree.value=a+(b-a)*rand(1);
        tr_left=train(train(:,tree.feat)<tree.value,:);
        tr_right=train(train(:,tree.feat)>=tree.value,:);
    end

    %Computing the proxy
    Nl=size(tr_left,1);
    [tree.proxy,lL,lR]=gini_oc_proxy(tree.limits,param.gamma,tree.volume,tree.feat,tree.value,N,Nl);

    %Setting data in the left child
    tree.left.height=tree.height+1;
    tree.left.volume=lL;
    tree.left.path=[tree.path '0'];
    tmp=tree.limits;
    tmp(tree.feat,2)=tree.value;
    tree.left.limits=tmp;
    tree.left=iTree(tr_left,tree.left,param);

    %Setting data in the right child
    tree.right.volume=lR;
    tree.right.path=[tree.path '1'];
    tmp=tree.limits;
    tmp(tree.feat,1)=tree.value;
    tree.right.limits=tmp;
    tree.right.height=tree.height+1;
    tree.right=iTree(tr_right,tree.right,param);
end
end